@@ -94,6 +94,18 @@ enum class type_init_flags : uint32_t {
 // See internals.h
 struct nb_alias_chain;
 
+// Implicit conversions for C++ type bindings, used in type_data below
+struct implicit_t {
+    const std::type_info **cpp;
+    bool (**py)(PyTypeObject *, PyObject *, cleanup_list *) noexcept;
+};
+
+// Forward and reverse mappings for enumerations, used in type_data below
+struct enum_tbl_t {
+    void *fwd;
+    void *rev;
+};
+
 /// Information about a type that persists throughout its lifetime
 struct type_data {
     uint32_t size;
@@ -111,24 +123,13 @@ struct type_data {
     void (*copy)(void *, const void *);
     void (*move)(void *, void *) noexcept;
     union {
-        // Implicit conversions for C++ type bindings
-        struct {
-            const std::type_info **cpp;
-            bool (**py)(PyTypeObject *, PyObject *, cleanup_list *) noexcept;
-        } implicit;
-
-        // Forward and reverse mappings for enumerations
-        struct {
-            void *fwd;
-            void *rev;
-        } enum_tbl;
+        implicit_t implicit;  // for C++ type bindings
+        enum_tbl_t enum_tbl;  // for enumerations
     };
     void (*set_self_py)(void *, PyObject *) noexcept;
     bool (*keep_shared_from_this_alive)(PyObject *) noexcept;
-#if defined(Py_LIMITED_API)
     uint32_t dictoffset;
     uint32_t weaklistoffset;
-#endif
 };
 
 /// Information about a type that is only relevant when it is being created
@@ -329,8 +330,18 @@ inline void *type_get_slot(handle h, int slot_id) {
 #endif
 }
 
+template <typename Visitor> struct def_visitor {
+  protected:
+    // Ensure def_visitor<T> can only be derived from, not constructed
+    // directly
+    def_visitor() {
+        static_assert(std::is_base_of_v<def_visitor, Visitor>,
+                      "def_visitor uses CRTP: def_visitor<T> should be "
+                      "a base of T");
+    }
+};
 
-template <typename... Args> struct init {
+template <typename... Args> struct init : def_visitor<init<Args...>> {
     template <typename T, typename... Ts> friend class class_;
     NB_INLINE init() {}
 
@@ -355,7 +366,7 @@ private:
     }
 };
 
-template <typename Arg> struct init_implicit {
+template <typename Arg> struct init_implicit : def_visitor<init_implicit<Arg>> {
     template <typename T, typename... Ts> friend class class_;
     NB_INLINE init_implicit() { }
 
@@ -402,7 +413,7 @@ namespace detail {
                 cpp_function_def(
                     [](handle type) {
                         if (!type_check(type))
-                            detail::raise_cast_error();
+                            throw cast_error();
                         return inst_alloc(type);
                     },
                     scope(cls), name("__new__"));
@@ -416,13 +427,46 @@ namespace detail {
             }
         }
     }
+
+    // Call policy that ensures __new__ returns an instance of the correct
+    // Python type, even when deriving from the C++ class in Python
+    struct new_returntype_fixup_policy {
+        static inline void precall(PyObject **, size_t,
+                                   detail::cleanup_list *) {}
+        NB_NOINLINE static inline void postcall(PyObject **args, size_t,
+                                                PyObject *&ret) {
+            handle type_requested = args[0];
+            if (ret == nullptr || !type_requested.is_type())
+                return; // somethign strange about this call; don't meddle
+            handle type_created = Py_TYPE(ret);
+            if (type_created.is(type_requested))
+                return; // already created the requested type so no fixup needed
+
+            if (type_check(type_created) &&
+                PyType_IsSubtype((PyTypeObject *) type_requested.ptr(),
+                                 (PyTypeObject *) type_created.ptr()) &&
+                type_info(type_created) == type_info(type_requested)) {
+                // The new_ constructor returned an instance of a bound type T.
+                // The user wanted an instance of some python subclass S of T.
+                // Since both wrap the same C++ type, we can satisfy the request
+                // by returning a pyobject of type S that wraps a C++ T*, and
+                // handling the lifetimes by having that pyobject keep the
+                // already-created T pyobject alive.
+                object wrapper = inst_reference(type_requested,
+                                                inst_ptr<void>(ret),
+                                                /* parent = */ ret);
+                handle(ret).dec_ref();
+                ret = wrapper.release().ptr();
+            }
+        }
+    };
 }
 
 template <typename Func, typename Sig = detail::function_signature_t<Func>>
 struct new_;
 
 template <typename Func, typename Return, typename... Args>
-struct new_<Func, Return(Args...)> {
+struct new_<Func, Return(Args...)> : def_visitor<new_<Func, Return(Args...)>> {
     std::remove_reference_t<Func> func;
 
     new_(Func &&f) : func((detail::forward_t<Func>) f) {}
@@ -442,17 +486,19 @@ struct new_<Func, Return(Args...)> {
               std::is_same_v<detail::intrinsic_t<Args>, kwargs>) + ... + 0);
         detail::wrap_base_new(cl, sizeof...(Args) > num_defaults + num_varargs);
 
-        auto wrapper = [func = (detail::forward_t<Func>) func](handle, Args... args) {
-            return func((detail::forward_t<Args>) args...);
+        auto wrapper = [func_ = (detail::forward_t<Func>) func](handle, Args... args) {
+            return func_((detail::forward_t<Args>) args...);
         };
 
+        auto policy = call_policy<detail::new_returntype_fixup_policy>();
         if constexpr ((std::is_base_of_v<arg, Extra> || ...)) {
             // If any argument annotations are specified, add another for the
             // extra class argument that we don't forward to Func, so visible
             // arg() annotations stay aligned with visible function arguments.
-            cl.def_static("__new__", std::move(wrapper), arg("cls"), extra...);
+            cl.def_static("__new__", std::move(wrapper), arg("cls"), extra...,
+                          policy);
         } else {
-            cl.def_static("__new__", std::move(wrapper), extra...);
+            cl.def_static("__new__", std::move(wrapper), extra..., policy);
         }
         cl.def("__init__", [](handle, Args...) {}, extra...);
     }
@@ -579,21 +625,9 @@ public:
         return *this;
     }
 
-    template <typename... Args, typename... Extra>
-    NB_INLINE class_ &def(init<Args...> &&arg, const Extra &... extra) {
-        arg.execute(*this, extra...);
-        return *this;
-    }
-
-    template <typename Arg, typename... Extra>
-    NB_INLINE class_ &def(init_implicit<Arg> &&arg, const Extra &... extra) {
-        arg.execute(*this, extra...);
-        return *this;
-    }
-
-    template <typename Func, typename... Extra>
-    NB_INLINE class_ &def(new_<Func> &&arg, const Extra &... extra) {
-        arg.execute(*this, extra...);
+    template <typename Visitor, typename... Extra>
+    NB_INLINE class_ &def(def_visitor<Visitor> &&arg, const Extra &... extra) {
+        static_cast<Visitor&&>(arg).execute(*this, extra...);
         return *this;
     }
 
@@ -800,18 +834,24 @@ public:
 };
 
 template <typename Source, typename Target> void implicitly_convertible() {
-    using Caster = detail::make_caster<Source>;
+    if constexpr (!std::is_same_v<Source, Target>) {
+        using Caster = detail::make_caster<Source>;
+        static_assert(
+            !std::is_enum_v<Target> || !detail::is_base_caster_v<Target>,
+            "implicitly_convertible(): 'Target' cannot be an enumeration "
+            "unless it is opaque.");
 
-    if constexpr (detail::is_base_caster_v<Caster>) {
-        detail::implicitly_convertible(&typeid(Source), &typeid(Target));
-    } else {
-        detail::implicitly_convertible(
-            [](PyTypeObject *, PyObject *src,
-               detail::cleanup_list *cleanup) noexcept -> bool {
-                return Caster().from_python(src, detail::cast_flags::convert,
-                                            cleanup);
-            },
-            &typeid(Target));
+        if constexpr (detail::is_base_caster_v<Caster>) {
+            detail::implicitly_convertible(&typeid(Source), &typeid(Target));
+        } else {
+            detail::implicitly_convertible(
+                [](PyTypeObject *, PyObject *src,
+                   detail::cleanup_list *cleanup) noexcept -> bool {
+                    return Caster().from_python(src, detail::cast_flags::convert,
+                                                cleanup);
+                },
+                &typeid(Target));
+        }
     }
 }
 
